import os
from typing import (
    Callable,
    Literal,
    Optional,
    TypeVar,
    overload,
)

from envtyped.repository import Repository, RepositoryEnv

T = TypeVar("T")
P = TypeVar("P", bound=type)


class Config:
    is_loaded = False

    def __init__(self, reposetory: Optional[Repository] = None) -> None:
        self.reposetory = reposetory or RepositoryEnv(".env")
        self.is_loaded = False

    @overload
    def get(
        self,
        key: str,
        default: P = ...,
        cast: None = ...,
        optional: Literal[False] = ...,
    ) -> str: ...

    @overload
    def get(
        self,
        key: str,
        default: P = ...,
        cast: None = ...,
        optional: Literal[True] = ...,
    ) -> Optional[str]: ...

    @overload
    def get(
        self,
        key: str,
        default: Optional[T] = ...,
        cast: Optional[Callable[[str], T]] = ...,
        optional: Literal[False] = ...,
    ) -> T: ...

    @overload
    def get(
        self,
        key: str,
        default: Optional[P] = ...,
        cast: Optional[P] = ...,
        optional: Literal[False] = ...,
    ) -> P: ...

    @overload
    def get(
        self,
        key: str,
        default: Optional[T] = ...,
        cast: Optional[Callable[[str], T]] = ...,
        optional: Literal[True] = ...,
    ) -> Optional[T]: ...

    def get(
        self,
        key: str,
        default: Optional[T | str] = None,
        cast: Optional[P | Callable[[str], T]] = None,
        optional: bool = False,
    ) -> Optional[T | P | str]:
        if not self.is_loaded:
            self.is_loaded = self.reposetory.load_config()

        value: Optional[T | str] = None

        if key in os.environ:
            value = os.environ[key]
        elif key in self.reposetory:
            value = self.reposetory[key]
        else:
            if default is not None:
                return default

            if not optional:
                raise KeyError(f"Key {key} not found in config or environment")

            return None

        assert value is not None

        if cast is not None:
            try:
                return cast(value)
            except Exception as e:
                raise ValueError(f"Cannot cast {value} to {cast}") from e

        return value


config = Config()



from envtyped.config import Config, config
from envtyped.helpers import Choice, Csv
from envtyped.repository import Repository, RepositoryEnv

__all__ = [
    "Choice",
    "Csv",
    "Config",
    "config",
    "Repository",
    "RepositoryEnv",
]



from pathlib import Path
from typing import (
    Dict,
    Optional,
    Protocol,
    TypeVar,
)

T = TypeVar("T")


class Repository(Protocol):
    def load_config(self) -> bool: ...

    def __getitem__(self, key: str) -> Optional[str]: ...

    def __contains__(self, key: str) -> bool: ...


class RepositoryEnv(Repository):
    def __init__(self, file_path: str) -> None:
        self.file_path = file_path
        self.config: Dict[str, str] = {}

    def load_config(self) -> bool:
        if not Path(self.file_path).exists():
            raise FileNotFoundError(f"File {self.file_path} not found")

        with open(self.file_path, "r") as f:
            for line in f:
                if line.startswith("#"):
                    continue

                key, value = line.strip().split("=")
                key = key.strip()
                value = value.strip()
                self.config[key] = value

        return True

    def __getitem__(self, key: str) -> Optional[str]:
        return self.config[key]

    def __contains__(self, key: str) -> bool:
        return key in self.config



from __future__ import annotations

import string
from shlex import shlex
from typing import (
    Callable,
    Generic,
    Optional,
    TypeVar,
    overload,
)

T = TypeVar("T")


class Csv(Generic[T]):
    def __init__(
        self,
        cast: Optional[type | Callable[[str], T]] = None,
        delimiter=",",
        strip=string.whitespace,
    ) -> None:
        self.cast = cast
        self.delimiter = delimiter
        self.strip = strip

    @overload
    def __call__(self: Csv[None], value: Optional[str]) -> list[str]: ...
    @overload
    def __call__(self, value: Optional[str]) -> list[T]: ...

    def __call__(self, value: Optional[str]) -> list[T] | list[str]:
        """The actual transformation"""
        if value is None:
            return []

        splitter = shlex(value, posix=True)
        splitter.escape = ""
        splitter.whitespace = self.delimiter
        splitter.whitespace_split = True

        if self.cast is None:
            return [s.strip(self.strip) for s in splitter]

        return [self.transform(s) for s in splitter]

    def transform(self, s: str) -> T:
        if self.cast is None:
            raise TypeError("Cast is None")

        return self.cast(s.strip(self.strip))


class Choice(Generic[T]):
    def __init__(self, choices: list[T]) -> None:
        self.choices = choices

    @overload
    def __call__(self: Choice[None], value: Optional[str]) -> str: ...
    @overload
    def __call__(self, value: Optional[str]) -> T: ...
    def __call__(self, value: Optional[str]) -> T | str:
        if value not in self.choices:
            raise ValueError(f"Value {value} not in choices {self.choices}")

        return value
